#ifndef ARKI_TYPES_H
#define ARKI_TYPES_H

/// arkimet metadata type system
#include <arki/defs.h>
#include <arki/core/fwd.h>
#include <arki/types/fwd.h>
#include <string>
#include <vector>
#include <memory>

struct lua_State;

namespace arki {
struct BinaryEncoder;
struct BinaryDecoder;
struct Emitter;
struct Formatter;

/// dynamic cast between two unique_ptr
template<typename B, typename A>
std::unique_ptr<B> downcast(std::unique_ptr<A> orig)
{
    if (!orig.get()) return std::unique_ptr<B>();

    // Cast, but leave ownership to orig
    B* dst = dynamic_cast<B*>(orig.get());

    // If we fail here, orig will clean it
    if (!dst)
    {
        std::string msg("cannot cast smart pointer from ");
        msg += typeid(A).name();
        msg += " to ";
        msg += typeid(B).name();
        throw std::runtime_error(std::move(msg));
    }

    // Release ownership from orig: we still have the pointer in dst
    orig.release();

    // Transfer ownership to the new unique_ptr and return it
    return std::unique_ptr<B>(dst);
}

/// upcast between two unique_ptr
template<typename B, typename A>
std::unique_ptr<B> upcast(std::unique_ptr<A> orig)
{
    return std::unique_ptr<B>(orig.release());
}

namespace emitter {
namespace memory {
struct Mapping;
}
}

namespace types {

// Parse name into a type code, returning TYPE_INVALID if it does not match
Code checkCodeName(const std::string& name);
// Parse name into a type code, throwing an exception if it does not match
Code parseCodeName(const std::string& name);
std::string formatCode(const Code& c);
static inline std::ostream& operator<<(std::ostream& o, const Code& c) { return o << formatCode(c); }

/**
 * Base class for implementing arkimet metadata types
 */
struct Type
{
	virtual ~Type() {}

    /**
     * Make a copy of this type. The caller will own the newly created object
     * returned by the function.
     *
     * This does not return an unique_ptr to allow to use covariant return types
     * in implementations. Use cloneType() for a version with explicit memory
     * ownership of the result.
     */
    virtual Type* clone() const = 0;

    /// Make a copy of this type
    std::unique_ptr<Type> cloneType() const { return std::unique_ptr<Type>(clone()); }

    /// Comparison (<0 if <, 0 if =, >0 if >)
    virtual int compare(const Type& o) const;
    bool operator<(const Type& o) const { return compare(o) < 0; }
    bool operator<=(const Type& o) const { return compare(o) <= 0; }
    bool operator>(const Type& o) const { return compare(o) > 0; }
    bool operator>=(const Type& o) const { return compare(o) >= 0; }

    /// Equality
    virtual bool equals(const Type& t) const = 0;
    bool operator==(const Type& o) const { return equals(o); }
    bool operator!=(const Type& o) const { return !equals(o); }

    /// Compare with a stringified version, useful for testing
    bool operator==(const std::string& o) const;

	/**
	 * Tag to identify this metadata item.
	 *
	 * This is used, for example, when creating error messages, or to identify
	 * the item when serialising to YAML.
	 */
	virtual std::string tag() const = 0;

	/// Serialisation code
	virtual types::Code type_code() const = 0;

	/// Length in bytes of the size field when serialising
	virtual size_t serialisationSizeLength() const = 0;

	/**
	 * Encoding to compact binary representation, without identification
	 * envelope
	 */
	virtual void encodeWithoutEnvelope(BinaryEncoder& enc) const = 0;

    /// Encode to compact binary representation, with identification envelope
    void encodeBinary(BinaryEncoder& enc) const;

    /// Encode to compact binary representation, with identification envelope
    std::vector<uint8_t> encodeBinary() const;

    /// Write as a string to an output stream
    virtual std::ostream& writeToOstream(std::ostream& o) const = 0;

    std::string to_string() const;

    /// Serialise using an emitter
    virtual void serialise(Emitter& e, const Formatter* f=0) const;
    virtual void serialiseLocal(Emitter& e, const Formatter* f=0) const = 0;

	/**
	 * Return a matcher query (without the metadata type prefix) that
	 * exactly matches this metadata item
	 */
	virtual std::string exactQuery() const;

    /// Push to the LUA stack a userdata with a copy of this item
    void lua_push(lua_State* L) const;

	/**
	 * Lookup members by name and push them in the Lua stack
	 *
	 * @return true if name matched a member and an element was pushed on
	 *         stack, else false
	 */
	virtual bool lua_lookup(lua_State* L, const std::string& name) const;

	/**
	 * Return the lua type name (i.e. arki.type.<something>)
	 * used to register the metatable for this type
	 */
	virtual const char* lua_type_name() const = 0;

	/**
	 * Given a metatable on top of the stack, register methods for this
	 * object on it
	 */
	virtual void lua_register_methods(lua_State* L) const;

    /**
     * Check that the element at \a idx is a Type userdata
     *
     * @return the Type element, or undefined if the check failed
     */
    static Type* lua_check(lua_State* L, int idx, const char* prefix = "arki.types");

    template<typename T>
    static T* lua_check(lua_State* L, int idx)
    {
        return dynamic_cast<T*>(lua_check(L, idx, traits<T>::type_lua_tag));
    }

    static void lua_loadlib(lua_State* L);

    /**
     * Return true of either both pointers are null, or if they are both
     * non-null and the two types test for equality.
     */
    static inline bool nullable_equals(const Type* a, const Type* b)
    {
        if (!a && !b) return true;
        if (!a || !b) return false;
        return a->equals(*b);
    }
    template<typename A, typename B>
    static inline bool nullable_equals(const std::unique_ptr<A>& a, const std::unique_ptr<B>& b)
    {
        return nullable_equals(a.get(), b.get());
    }
    /**
     * Return the comparison value of a and b, assuming that a null pointer
     * tests smaller than any type.
     */
    static inline int nullable_compare(const Type* a, const Type* b)
    {
        if (!a && !b) return 0;
        if (!a) return -1;
        if (!b) return 1;
        return a->compare(*b);
    }
    template<typename A, typename B>
    static inline bool nullable_compare(const std::unique_ptr<A>& a, const std::unique_ptr<B>& b)
    {
        return nullable_compare(a.get(), b.get());
    }
};

/// Write as a string to an output stream
inline std::ostream& operator<<(std::ostream& o, const Type& t)
{
    return t.writeToOstream(o);
}



template<typename BASE>
struct CoreType : public Type
{
    types::Code type_code() const override { return traits<BASE>::type_code; }
    size_t serialisationSizeLength() const override { return traits<BASE>::type_sersize_bytes; }
    std::string tag() const override { return traits<BASE>::type_tag; }
    const char* lua_type_name() const override { return traits<BASE>::type_lua_tag; }
    static void lua_loadlib(lua_State* L);
};

template<typename BASE>
struct StyledType : public CoreType<BASE>
{
	typedef typename traits<BASE>::Style Style;

	// Get the element style
	virtual Style style() const = 0;

    // Default implementations of Type methods
    void encodeWithoutEnvelope(BinaryEncoder& enc) const override;
    int compare(const Type& o) const override;
    virtual int compare_local(const BASE& o) const { return style() - o.style(); }

    virtual void serialiseLocal(Emitter& e, const Formatter* f=0) const;

	virtual bool lua_lookup(lua_State* L, const std::string& name) const;

    static Style style_from_mapping(const emitter::memory::Mapping& m);
};


/**
 * Decode an item encoded in binary representation with envelope, from a
 * decoder
 */
std::unique_ptr<Type> decode(BinaryDecoder& dec);

std::unique_ptr<Type> decodeInner(types::Code, BinaryDecoder& dec);
std::unique_ptr<Type> decodeString(types::Code, const std::string& val);
std::unique_ptr<Type> decodeMapping(const emitter::memory::Mapping& m);
/// Same as decodeMapping, but does not look for the item type in the mapping
std::unique_ptr<Type> decodeMapping(types::Code, const emitter::memory::Mapping& m);
std::string tag(types::Code);


/**
 * Read a data bundle from a POSIX file descriptor
 */
struct Bundle
{
    /// Bundle signature
    std::string signature;
    /// Bundle version
    unsigned version;
    /// Data length
    size_t length;
    /// Bundle data
    std::vector<uint8_t> data;

    /**
     * Read only the bundle header
     *
     * @return true if a bundle header was read, false on end of file
     */
    bool read_header(core::NamedFileDescriptor& fd);

    /**
     * Read the bundle data after read_header has been called
     *
     * @return true if all bundle data was read, false on end of file
     */
    bool read_data(core::NamedFileDescriptor& fd);

    /**
     * read_header and read_data together
     *
     * @return true if all was read, false on end of file
     */
    bool read(core::NamedFileDescriptor& fd);
};

}

}

#endif
